/* grafgen : Precision graphical genotypes based on marker data */
/* Authors : B. Servin, C. Dillmann, G. Decoux, and F. Hospital */
/* Institution :
   Station de Genetique Vegetale INRA / UPS / INAPG 
   91190 Gif-sur-Yvette - FRANCE */

/*     This file is part of Grafgen. */

/*     Grafgen is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or */
/*     (at your option) any later version. */

/*     Grafgen is distributed in the hope that it will be useful, */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*     GNU General Public License for more details. */

/*     You should have received a copy of the GNU General Public License */
/*     along with Grafgen; if not, write to the Free Software */
/*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



#include "grafgen.h"

#include <assert.h>
#include <gd.h>
#include <gdfontt.h>
#include <gdfonts.h>
#include <gdfontmb.h>
#include <gdfontl.h>
#include <gdfontg.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#define LONG 256
#define GLperC 5		/* max number of linkage group for a chromosome */
#define NGL GLperC*CHROMO

char ficod[2048];		/* name of the file containing the coding system */

double posx;			/* position of the additional locus */
double xstep;			/* step to use for scanning */
double x_beg;			/* beginning of scan */
double x_end;			/* end of scan */

int mstep;			/* nb of markers to be taken into account on *one* side of posx */

char gl_name[NGL][LONG];
char mk_name[NGL][LOCUS][LONG];	/* names of (marker) loci */
double mk_pos[NGL][LOCUS + 1];	/* Positions of marker loci on the
				   chromosome */

/* Structure used to store info on chromosomes and linkage groups */
struct
{
  int nb;  /* Number of linkage groups */
  int index[NGL]; /* Which linkage group they are */
} lg[CHROMO];   
int endc[NGL];			/* index of the last (marker) locus of linkage groupes */
double lenc[NGL];		/* position of the last (marker) locus on the 
				   linkage group = LG length */


int geno_mat[NGL][GENO][LOCUS + 1];	/* Observed genotypes (OGs) of the
					   maternal ancestors */
int geno_pat[NGL][GENO][LOCUS + 1];	/* Observed Genotypes of 
					   paternal ancestors (OGs)
					   when an hybrid cross or a backcross
					   is performed */
int geno_off[NGL][GENO][LOCUS + 1];	/* Observed genotypes of the 
					   offspring (OGs) */

struct
{
  double th;			/* Threshold s.a. if p>th : plot */
  char color[2048];		/* color to plot */
} threshold[N];

int addloc_mat[GENO];		/* Observed Genotypes of 
				   maternal ancestors (OGs) 
				   for the additional locus */
int addloc_pat[GENO];		/* Observed Genotypes of 
				   paternal ancestors (OGs)
				   (in case of hybridization or
				   backcrossing) for the 
				   additional locus */
int addloc_off[GENO];		/* Observed Genotypes (OGs) 
				   to allocate to the additional locus 
				   for the offspring */
char mating_system[GENO][256];	/* Mating systems performed between
				   generations */
int locus_tot, cod_tot;

int nb_addloc = 0;		/* number of addloc_off genotypes */

int maxall[] = { 0, 0, 7, 4, 3 };	/* number of loci that can be handled
					   together for the computation for 0,
					   1, 2, 3, 4 alleles per locus, 
					   respectively */
int homo, hetero;

int gen;			/* number of generations in the breeding scheme */

int nb_off;			/* number of individuals in the offspring */
int nb_chr;			/* number of chromosomes */
int nb_mk=0;			/* overall number of (marker) loci on the genome */
int off_beg;			/*  first individual to plot */
int off_end;			/* last individual to plot ?? */

int chr_beg;			/* first chromosome to be taken into account for scan */

double probcomm[GENO];		/* frequencies of genotypes (markers
				   loci + additional locus) of the
				   maternal ancestors */
double probdiff[GENO][N];	/* frequencies of genotypes (marker loci
				   + additional locus) of the offspring */
double probcomm_mq[GENO];	/* frequencies of marker-genotypes of
				   the ancestors */
double probdiff_mq[GENO][N];	/* frequencies of marker genotypes of
				   the offspring */
double results[GENO][POSITIONS];

static int GENOME;		/* scan on the whole genome if 1 */
static int INTER;		/* scan on a segment of chromosome if 1 */
static int VERBOSE=0;

/* GRAPHICS DEFINITIONS */
char GFORMAT[4];		/* format of graphics output */
int type_disp, all_disp;	/* type of display and allele to display */
const int Xsize=150;
const int TitleSize=30;
const int scale=2;
int Ysize;		/* size of the chromosome sub-picture */
void (*calcol) (double r);	/* functionto use for graphical display */
char title[2048];
char fic_thresh[2048];		/* Name of the file containings thresholds values */
int YSIZE;
double expect_val;		/* expected value for the mean allele frequency */

/***********************************************************************/

/* reads the first entry of a string 'st' with format 'format',
   ignoring "non-informative" elements such as blank spaces, tabulations, 
   ends of line ... and puts results in res, the pointer on string 'st' 
   is increased in order to point on 
   the following non-space element. Thus, successive calls to 'reading' allow 
   to read the successive non-space elements of the string 'st'*/

int
reading (st, format, res)
     char **st, *format;
     void *res;
{
  int i;
  char *rs;
  rs = *st;
  for (i = 0; isspace (rs[i]); i++);
  if (!rs[i])
    return 0;
  for (; !isspace (rs[i]); i++);
  if (rs[i])
    rs[i++] = 0;
  if (!sscanf (*st, format, res))
    return 0;
  *st += i;
  return 1;
}

/* */


/* reads the file containing the coding system that defines relationships 
   between OGs and TGs */
/* The relationships are stored in the structure 'coding_system' 
   defined in the header file mdm.h */

static void
reading_coding_system (file)
     char *file;
{
  FILE *fd;
  char ligne[2048], *s;
  char temp[2048];
  int i, j, cas;

  if ((fd = fopen (file, "r")) == NULL)
    {
      fprintf (stderr, "cannot read %s\n", file);
      exit (1);
    }
  for (i = 0; i < P; i++)
    {
      coding_system[i].len = 0;
      for (j = 0; j < N; j++)
	{
	  coding_system[i].possible[j][0] = coding_system[i].possible[j][1] =
	    0;
	}
    }
  nball = -1;
  while (fgets (ligne, 2047, fd))
    {
      s = ligne;
      if (!reading (&s, "%d", &cas))
	continue;
      if (cas < 0)
	{
	  fclose (fd);
	  fprintf (stderr, "invalid value in the coding system : %s\n", file);
	  exit (1);
	}
      if (nball == -1)
	{
	  if (cas < 2 || cas > 4)
	    {
	      fclose (fd);
	      fprintf (stderr,
		       "invalid number of alleles in the coding system : %s\n",
		       file);
	      exit (1);
	    }
	  nball = cas;
	  continue;
	}
      while (reading (&s, "%s", temp))
	{
	  if (sscanf (temp, "%d/%d", &i, &j) != 2)
	    {
	      fclose (fd);
	      fprintf (stderr, "invalid format in the coding system : %s\n",
		       file);

	      exit (1);
	    }
	  if ((i < 0 || i >= nball) || (j < 0 || j >= nball))
	    {
	      fclose (fd);
	      fprintf (stderr,
		       "invalid allele value in the coding system : %s\n",
		       file);
	      exit (1);
	    }
	  coding_system[cas].len++;
	  if (coding_system[cas].len > N)
	    {
	      fclose (fd);
	      fprintf (stderr,
		       "two many values for the OG %d in the coding system : %s\n",
		       cas, file);
	      exit (1);
	    }
	  coding_system[cas].possible[coding_system[cas].len - 1][0] = i;
	  coding_system[cas].possible[coding_system[cas].len - 1][1] = j;
	}
    }
  fclose (fd);
  cod_tot = cas;
}

/* */

void
codsys_analysis ()
{

  int hetero_len;
  /* hetero_len is the number of TG which corresponds to the wanted 
     heterozygote OG:
     ex nball = 2; wanted OG <=> 0/1 1/0 : length 2
     ex nball = 3; wanted OG <=> 0/1 1/0 0/2 2/0 : length 4 */
  int c, tg, aa;
  int indice, mark;


  hetero_len = (nball - 1) * 2;
  homo = hetero = -1;

  /* check if wanted OGs exist */
  for (c = 0; c < N; c++)
    {
      if (coding_system[c].len == 0)
	continue;
      /* is this the wanted homozygote ? */
      if ((homo == -1) &&
	  (coding_system[c].len == 1) &&
	  (coding_system[c].possible[0][0] == all_disp) &&
	  (coding_system[c].possible[0][0] ==
	   coding_system[c].possible[0][1]))
	{
	  homo = c;
	  continue;
	}
      /* is this the wanted heterozygote ? */
      if ((hetero == -1) && (coding_system[c].len == hetero_len))
	{
	  tg = 0;
	  while (tg < hetero_len)
	    {
	      /* tg = homozygote : not good */
	      if (coding_system[c].possible[tg][0] ==
		  coding_system[c].possible[tg][1])
		{
		  hetero = -1;
		  break;
		}
	      /* not a single all_disp in this tg : not good */
	      if ((coding_system[c].possible[tg][0] != all_disp) &&
		  (coding_system[c].possible[tg][1] != all_disp))
		{
		  hetero = -1;
		  break;
		}
	      /* a priori ok for this tg */
	      hetero = c;
	      /* see next one */
	      tg++;
	    }
	}
    }
  /* case homo is not there */
  if (homo == -1)
    {
      c = 0;
      /* find an unused OG code */
      while (coding_system[c].len)
	c++;
      /* this is the homozygote */
      coding_system[c].len = 1;
      coding_system[c].possible[0][0] = coding_system[c].possible[0][1] =
	all_disp;
      homo = c;
    }
  /* case hetero is not there */
  if (hetero == -1)
    {
      c = 0;
      indice = 0;
      /* find an unused OG code */
      while (coding_system[c].len)
	c++;
      /* this is the heterozygous */
      coding_system[c].len = hetero_len;
      /* puting all_disp where it belongs */
      for (tg = 0; tg < hetero_len; tg++)
	{
	  coding_system[c].possible[tg][indice] = all_disp;
	  indice = 1 - indice;
	}
      indice = 1 - indice;
      mark = 0;
      /* putting the other alleles in the holes */
      for (aa = 0; aa < nball; aa++)
	{
	  if (aa == all_disp)
	    continue;
	  for (tg = mark; tg < (mark + 2); tg++)
	    {
	      coding_system[c].possible[tg][indice] = aa;
	      indice = 1 - indice;
	    }
	  mark += 2;
	}
      hetero = c;
    }
}


static void
reading_threshold (file)
     char *file;
     /* function to read the threshold file */

{
  FILE *fd;
  char line[2048], *s;
  int i;
  int g, gg, ll;

  if ((fd = fopen (file, "r")) == NULL)
    {
      fprintf (stderr, "cannot read the file: %s\n", file);
      exit (1);
    }

  ll = g = 0;

  /*init threshold s.a. color = none, th = -1 */
  for (i = 0; i < N; i++)
    {
      threshold[i].th = 0;
      strncpy (threshold[i].color, "none", 4 * sizeof (char));
    }

  while (fgets (line, 2047, fd))
    {
      ll++;
      if (!strncmp (line, "#", 1))
	continue;
      g++;
      s = line;
      if (!reading (&s, "%d", &gg))
	{
	  fprintf (stderr, "Error in file %s line %d\n", file, ll);
	  exit (1);
	}
      if (!reading (&s, "%lf", &threshold[gg].th))
	{
	  fprintf (stderr, "Error in file %s line %d\n", file, ll);
	  exit (1);
	}
      if (!reading (&s, "%s", threshold[gg].color))
	{
	  fprintf (stderr, "Error in file %s line %d\n", file, ll);
	  exit (1);
	}
    }

  for (i = 0; i < nb_addloc; i++)
    {
      if (threshold[addloc_off[i]].th <= 0)
	printf ("Note: %d \tnot plotted\n", i);
    }
}

static int
reading_input (file)
     char *file;
     /* the function reading_input reads the infile */
{
  FILE *fd;
  char line[2048], buf[2048], *s, gl[10], cur_gl[10];
  int i, c, g;
  int cur_chrom;
  int chrom, mk;
  int nb_gl=0;

  for (c = 0; c < NGL; c++)
    {
      endc[c] = -1;
    }
  nb_chr = 0;

  if ((fd = fopen (file, "r")) == NULL)
    {
      fprintf (stderr, "cannot read %s\n", file);
      exit (1);
    }

  /* reads the number of generations of the breeding scheme */

  fgets (line, 2047, fd);
  if (!strncmp (line, "GENER =", 7))
    {
      if (sscanf (line, "GENER = %d", &gen) != 1)
	{
	  fclose (fd);
	  fprintf (stderr,
		   "Invalid number of generation in the main input file :%s\n",
		   file);
	}
      if (gen > GENO)
	{
	  fprintf (stderr, "Do you real mean %d generations ??????\n", gen);
	  exit (1);
	}
    }
  else
    {
      fclose (fd);
      fprintf (stderr,
	       "The number of generation is missing in the file :%s\n", file);
      exit (1);
    }

  /* reads the number of individuals in the offpsring */

  fgets (line, 2047, fd);
  if (!strncmp (line, "OFFSPRING =", 11))
    {
      if (sscanf (line, "OFFSPRING = %d", &nb_off) != 1)
	{
	  fclose (fd);
	  fprintf (stderr, "invalid number of offspring genotypes : %d\n",
		   nb_off);
	  exit (1);
	}
      if (nb_off > GENO)
	{
	  fprintf (stderr, "Too many offsprings : %d, max number is : %d\n",
		   nb_off, GENO);
	  fclose (fd);
	  exit (1);
	}
    }
  else
    {
      fclose (fd);
      fprintf (stderr,
	       "The number of offspring individuals is missing in the file :%s\n",
	       file);
      exit (1);
    }

  /* reads the number of genotypes to be evaluated for the additional 
     locus in the offspring */

  fgets (line, 2047, fd);
  if (!strncmp (line, "EVAL", 4))
    {
      if (sscanf (line, "EVALUATED GENOTYPES = %d", &nb_addloc) != 1)
	{
	  fclose (fd);
	  fprintf (stderr, "invalid number of offspring genotypes : %d\n",
		   nb_addloc);
	  exit (1);
	}
      if (nb_addloc > N)
	{
	  fprintf (stderr, "You don't want to evaluate %d genotypes :)\n",
		   nb_addloc);
	  fclose (fd);
	  exit (1);
	}
    }
  else
    {
      fclose (fd);
      fprintf (stderr,
	       "The number of offspring individuals is missing in the file :%s\n",
	       file);
      exit (1);
    }

  /* reads the name of the file containing the coding system */

  fgets (line, 2047, fd);
  if (!strncmp (line, "COD", 3))
    {
      if (sscanf (line, "CODING SYSTEM FILE = %s", ficod) != 1)
	{
	  fprintf (stderr, "%s\n", ficod);
	  fclose (fd);
	  fprintf (stderr,
		   "Missing path to coding system file in the file : %s\n",
		   file);
	  exit (1);
	}
    }
  else
    {
      fclose (fd);
      fprintf (stderr, "Missing path to coding system file in the file :%s\n",
	       file);
      exit (1);
    }
  fgets (line, 2047, fd);
  fgets (line, 2047, fd);
  if (strncmp (line, "MATING", 6))
    {
      fprintf (stderr,
	       "The mating system information is missing, or the keywords are absent : stopping\n");
      exit (1);
    }
  fgets (line, 2047, fd);

  /* reads the mating systems used between each generation of the 
     breeding scheme */
  for (i = 0; i < gen; i++)
    {
      fgets (line, 2047, fd);
      s = line;
      reading (&s, "%s", buf);
      reading (&s, "%s", mating_system[i]);
    }

  fgets (line, 2047, fd);
  fgets (line, 2047, fd);
  if (strncmp (line, "ADD", 3))
    {
      fprintf (stderr,
	       "The additional locus information is missing, or the keywords are absent : stopping\n");
      exit (1);
    }
  fgets (line, 2047, fd);


  for (g = 0; g < gen; g++)
    {
      fgets (line, 2047, fd);
      s = line;
      if (!strncmp (line, "none", 4))
	{
	  fprintf (stderr, "Additional locus information missing !\n");
	  exit (1);
	}
      reading (&s, "%s", buf);
      reading (&s, "%d", &addloc_mat[g]);
      if (!strncmp (mating_system[g], "hyb", 3)
	  || !strncmp (mating_system[g], "bc", 2))
	{
	  reading (&s, "%d", &addloc_pat[g]);
	}
    }
  fgets (line, 2047, fd);
  s = line;
  /* reads the genotype to allocate to the additional locus in 
     the offspring */
  reading (&s, "%s", buf);
  for (g = 0; g < nb_addloc; g++)
    {
      reading (&s, "%d", &addloc_off[g]);
    }

  fgets (line, 2047, fd);
  fgets (line, 2047, fd);
  if (strncmp (line, "GENETIC", 7))
    {
      fprintf (stderr,
	       "The GENETIC MAP AND MARKER GENOTYPES section is missing or keywords are absent : stopping\n");
      exit (1);
    }
  fgets (line, 2047, fd);
  fgets (line, 2047, fd);
  if (strncmp (line, "CHROM", 5))
    {
      fprintf (stderr,
	       "The heading line for loci information is missing or keywords are absent : stopping\n");
      exit (1);
    }

  /* for each line, reads : the chromosome to which belongs the (marker) 
     locus, the name of the marker locus, 
     its position on the chromosome, the genotypes of the ancestors 
     (from the first generation to the last),
     and the genotypes of each offspring */
  /* initialisation */ 
  mk = 0; /* Indexes the number of markers on the current linkage group */
  cur_gl[0]='\0'; /* Name of the linkage group considered */
  cur_chrom=-1;

  while (fgets (line, 2047, fd))
    {
      if (!strncmp (line, "\n", 1))
	continue;
      if (!strncmp (line, "#", 1))
	continue;
      s = line;

      /* First read the chromosome information */
      reading (&s, "%s", buf);
      /*  This is a chromosome with more than one linkage groups */
      if (sscanf (buf, "%d%s", &chrom, gl) == 2)
	{
	  chrom--;
	  if (chrom < 0) {
	    fprintf(stderr,"Error *** Chromosome index should start at 1\n");
	    exit(EXIT_FAILURE);
	  }
	  if (chrom < cur_chrom) {
	    fprintf(stderr,"Error *** The chromosomes / markers must be sorted\n");
	    exit(EXIT_FAILURE);
	  }
	  
	  if (cur_chrom != chrom) {
	    /* This is a new chromosome */
	    /* Hence a new linkage group also*/
	    strcpy(gl_name[nb_gl], buf);
	    lg[chrom].index[0] = nb_gl;
	    lg[chrom].nb=1;
	    mk = 0;
	    cur_chrom = chrom;
	    strcpy(cur_gl,gl);
	    nb_gl++;
	  }
	  else if( strcmp(cur_gl, gl) != 0) {
	    /* Same chromosome, new linkage group */
	    strcpy(gl_name[nb_gl], buf);
	    lg[chrom].index[lg[chrom].nb]=nb_gl;
	    lg[chrom].nb++;
	    strcpy(cur_gl,gl);
	    mk=0;
	    nb_gl++;
	  }
	  /* if this is not a new linkage group, we're done */
	}
      /* This chromosome has no linkage group information */
      else if (sscanf (buf, "%d", &chrom) == 1)
	{
	  chrom--;
	  if (cur_chrom != chrom) 
	    /* new chromosome */
	    {
	      mk = 0;
	      cur_chrom = chrom;
	      sprintf (gl_name[nb_gl], "%d", chrom+1);
	      cur_gl[0]='\0'; /* we must reset the GL information */
	      lg[chrom].index[0] = nb_gl;
	      lg[chrom].nb = 1;
	      nb_gl++;
	    }
	}
      else
	{
	  fprintf (stderr, "Unrecognized format for the chromosome index\n");
	  exit (1);
	}
      if (nb_gl > NGL)
	{
	  fprintf (stderr, "Too many chromosome to analyse %d > %d\n", nb_gl,
		   NGL);
	}
      /* read marker name */
      reading (&s, "%s", mk_name[nb_gl - 1][mk]);
      /* read marker pos */
      reading (&s, "%lf", &mk_pos[nb_gl - 1][mk]);
      /* read genotype data */
      for (g = 0; g < gen; g++)
	{
	  reading (&s, "%d", &geno_mat[nb_gl - 1][g][mk]);
	  if (!strncmp (mating_system[g], "hyb", 3)
	      || !strncmp (mating_system[g], "bc", 2))
	    {
	      reading (&s, "%d", &geno_pat[nb_gl - 1][g][mk]);
	    }
	}

      for (g = 0; g < nb_off; g++)
	{
	  reading (&s, "%d", &geno_off[nb_gl - 1][g][mk]);
	}

      if (mk > LOCUS)
	{
	  fprintf (stderr, "Too many loci on chromosome %s : %d > %d\n",
		   gl_name[nb_gl - 1], mk, LOCUS);
	  fclose (fd);
	  exit (1);
	}
      endc[nb_gl - 1] = mk;
      lenc[nb_gl - 1] = mk_pos[nb_gl - 1][mk];
      mk++;
      nb_mk++;
    }
  fclose (fd);
  nb_chr=chrom+1;
  return nb_gl;
}

/* OPTIONS */

void
print_help (char *argv[])
{
  fprintf (stderr, "\n");
  fprintf (stderr,
	   "*********************************************************************\n");
  fprintf (stderr,
	   "*                             GRAFGEN                               *\n");
  fprintf (stderr,
	   "*  Precise analysis of the genomic composition in inbred pedigrees  *\n");
  fprintf (stderr,
	   "*********************************************************************\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "Usage : \n   %s [options] infile \n", "grafgen");
  fprintf (stderr, "\n");
  fprintf (stderr,
	   "Parameters : \ninfile \t\t name of the file containing the input \n\n");
  fprintf (stderr, "Computation Options :\n---------------------\n");
  fprintf (stderr, " -h\t\t\t\tprint this help and exits\n");
  fprintf (stderr, " -g\t\t\t\tscan the whole genome (default)\n");
  fprintf (stderr,
	   " -c <index> (integer)    \tscan the chromosome defined by index\n");
  fprintf (stderr,
	   " -x <step>  (double)     \tdefine the step used for the scan\n");
  fprintf (stderr,
	   " -b <position> (double)  \tdefine the position where to begin the scan \n");
  fprintf (stderr, "                         \t(use with -c only)\n");
  fprintf (stderr,
	   " -e <position> (double)  \tdefine the position where to end the scan \n");
  fprintf (stderr, "                         \t(use with -c only)\n");
  fprintf (stderr,
	   " -m <step>  (integer)    \tdefine the number of markers (on one side of \n");
  fprintf (stderr,
	   "                         \tthe position) to be taken into account\n");
  fprintf (stderr,
	   " -i <indiv> (integer)    \tdefine the individual for which the PGG is\n");
  fprintf (stderr, "                         \tto plot\n");
  fprintf (stderr,
	   " -M <mean>  (double)     \texpected allele frequency in the population\n");
  fprintf (stderr, "Graphics Options :\n------------------\n");
  fprintf (stderr, " -f <jpg or png>         \tformat of graphics output\n");
  fprintf (stderr,
	   " -T <type>  (integer)    \ttype of display : none (0), genetic map (1),\n");
  fprintf (stderr,
	   "                         \t probability (2), allele dose (3),\n");
  fprintf (stderr,
	   "                         \tmean allele dose (4), threshold representation\n");
  fprintf (stderr,
	   "                         \tof the different genotypes (5)\n");
  fprintf (stderr,
	   " -t <arg> (integer)      \tadditional argument for -T option\n");
  fprintf (stderr, "                         \trequired for <type> => 2.\n");
  fprintf (stderr,
	   " -P <Title> (string)     \tname of the project to use as a prefix\n");
  fprintf (stderr, "                         \tfor output files\n");
  fprintf (stderr,
	   " -d <color type> (string)\tType of coloring for the output (bw, hot)\n");
}

void
options (int argc, char *argv[])
{
  int c;

  GENOME = 1;
  INTER = 0;

  /* Initialize Variables to default or testing values */
  x_beg = off_beg = chr_beg = 0;
  xstep = mstep = type_disp = 1;
  x_end = off_end = all_disp = -1;
  expect_val = -1;
  strncpy (GFORMAT, "def", 3 * sizeof (char));
  strncpy (title, "pgg", 3 * sizeof (char));
  calcol = heat;
  while (1)
    {
      c = getopt (argc, argv, "hgc:x:b:e:m:f:t:a:T:P:d:i:M:v");
      if (c == -1)
	{
	  break;
	}
      switch (c)
	{
	case 'h':
	  print_help (argv);
	  exit (1);
	case 'g':
	  GENOME = 1;
	  break;
	case 'v':
	  VERBOSE=1;
	  break;
	case 'c':
	  GENOME = 0;
	  chr_beg = atoi (optarg);
	  chr_beg--;
	  break;
	case 'x':
	  xstep = atof (optarg);
	  if (xstep <= 0)
	    {
	      fprintf (stderr, "Error : option '-x' : step for scan = %.2f\n",
		       xstep);
	      exit (1);
	    }
	  break;
	case 'b':
	  INTER = 1;
	  x_beg = atof (optarg);
	  if (x_beg < 0)
	    {
	      fprintf (stderr,
		       "Error : option '-b' : starting point have a negative coordinate %.3f\n",
		       x_beg);
	      exit (1);
	    }
	  break;
	case 'e':
	  INTER = 1;
	  x_end = atof (optarg);
	  if (x_end < 0)
	    {
	      fprintf (stderr,
		       "Error : option '-e' : ending point have a negative coordinate %.3f\n",
		       x_end);
	      exit (1);
	    }
	  break;
	case 'm':
	  mstep = atoi (optarg);
	  if (mstep == 0)
	    {
	      fprintf (stderr,
		       "Error : option '-m' : marker step must be > 0\n");
	      exit (1);
	    }
	  break;
	case 'f':
	  strncpy (GFORMAT, optarg, 4 * sizeof (char));
	  if (strncmp (optarg, "png", 3))
	    {
	      if (strncmp (optarg, "jpg", 3))
		{
		  fprintf (stderr,
			   "Error : option '-f' : Output format must be either 'png' or 'jpg'\n");
		  exit (1);
		}
	    }
	  break;
	case 'T':
	  type_disp = atoi (optarg);
	  if (type_disp == 4)
	    calcol = bluewhitered;
	  if (type_disp > 5)
	    {
	      fprintf (stderr, "Error : option '-T' : Value %d unavailable\n",
		       type_disp);
	      exit (1);
	    }
	  break;
	case 't':
	  if (type_disp < 2)
	    fprintf (stderr,
		     "Note: you do not need the '-t' sub-option with type %d\n",
		     type_disp);
	  else if (type_disp < 5)
	    all_disp = atoi (optarg);
	  else if (type_disp == 5)
	    {
	      strncpy (fic_thresh, optarg, 2048 * sizeof (char));
	    }
	  break;
	case 'P':
	  strncpy (title, optarg, 2048 * sizeof (char));
	  break;
	case 'd':
	  if (!strcmp (optarg, "hot"))
	    {
	      calcol = heat;
	    }
	  else if (!strcmp (optarg, "bw"))
	    calcol = BW;
	  else
	    {
	      fprintf (stderr,
		       "Error : option '-d' : coloring function = %s does not exist\n",
		       optarg);
	      exit (1);
	    }
	  break;
	case 'i':
	  off_beg = atoi (optarg);
	  if (off_beg <= 0)
	    {
	      fprintf (stderr,
		       "Error : option '-i' : negative individual index\n");
	      exit (1);
	    }
	  off_beg--;
	  off_end = off_beg + 1;
	  break;
	case 'M':
	  expect_val = atof (optarg);
	  calcol = bluewhitered; /* No choice here */
	  break;
	default:
	  print_help (argv);
	  exit (1);
	}
    }

  if (GENOME && INTER)
    {
      fprintf (stderr,
	       "You must input a chromosome on which the segment is to scan\n");
      exit (1);
    }
  if (type_disp > 1 && type_disp < 5 && all_disp < 0)
    {
      fprintf (stderr,
	       "Error : More Precision needed for option '-T %d' (use option -t)\n ",
	       type_disp);
      exit (1);
    }
  if (!strncmp (title, "pgg", 3))
    {
      fprintf (stderr, "Default prefix will be used for output files : %s\n",
	       title);
    }
}

/* END OPTIONS */

gdImagePtr
ImageCreateTrueColorWhiteBG (X, Y)
     int X, Y;

{

  gdImagePtr im;
  int white;

  im = gdImageCreateTrueColor (X, Y);
  white = gdImageColorAllocate (im, 255, 255, 255);
  gdImageFilledRectangle (im, 0, 0, X, Y, white);

  return im;
}

gdImagePtr
make_chromosome_image (chr, ppf, y_beg, y_end, indiv, p, r, ip)
     double p[POSITIONS];	/* scanned positions */
     double r[GENO][POSITIONS];	/* results to draw */
     int ip;
     int chr;			/* chromosome to draw */
     gdFontPtr ppf;
     double y_beg, y_end;
     int indiv;			/* offspring to draw */
{
  gdImagePtr im_chr;
  gdFontPtr gl_title;
  int gl_title_len;
  int i;
  /* indexes of the first and last marker to draw */
  int mk_beg, mk_end;

  /* lower existing text position */
  int poz_bottom[LOCUS];

  /* colour indexes */
  int fgcolor;
  int pggcolor;

  /* chromosome coordinates */
  int x1_chr, x2_chr, y1_chr, y2_chr;
  /* PGG coordinates */
  int y1_pgg, y2_pgg;
  /* marker coordinates */
  int x1_mk, x2_mk, y_mk[LOCUS], y_mk_name[LOCUS], x1_mk_name, x2_mk_name;

  /* string to store positions */
  char pozit[10];

  x1_chr = (int) (0.45 * Xsize);
  x2_chr = (int) (0.55 * Xsize);
  /* lines X coordinates */
  x1_mk = x1_chr - 5;
  x2_mk = x2_chr + 5;
  /* name X coordinates */
  x1_mk_name = x1_mk - 10;
  x2_mk_name = x2_mk + 10;

  y1_chr = 10;
  y2_chr = (int) (y1_chr + (y_end - y_beg) * scale);

  /* identify the indexes of first and last marker to draw */
  for (i = 0; (i < endc[chr]) && (mk_pos[chr][i] < y_beg); i++);
  mk_beg = i;
  for (i = 0; (i < endc[chr]) && mk_pos[chr][i] < y_end; i++);

  mk_end = i + 1;

  /* draw the markers and put names */
  for (i = mk_beg; i < mk_end; i++)
    {
      poz_bottom[i] = (i > mk_beg) ? poz_bottom[i - 1] : 0;
      /* line Y coordinate */
      y_mk[i] = (int) (y1_chr + (mk_pos[chr][i] - y_beg) * scale);
      y_mk_name[i] = (int) (y_mk[i] - 0.5 * ppf->h);
      if (poz_bottom[i] + 2 > y_mk_name[i])
	y_mk_name[i] = (int) (poz_bottom[i] + ppf->h);
      poz_bottom[i] = (int) (y_mk_name[i] + 0.5 * ppf->h);
    }

  gl_title = gdFontMediumBold;
  gl_title_len = strlen (gl_name[chr]);
  gl_title_len *= gl_title->w;

  Ysize = poz_bottom[mk_end - 1] + (2 * gl_title->h) + 10;	/* (last marker pos + title size + margin )  */

  YSIZE = (YSIZE < Ysize) ? Ysize : YSIZE;

  im_chr = ImageCreateTrueColorWhiteBG (Xsize, Ysize);
  fgcolor = gdImageColorAllocate (im_chr, 0, 0, 0);
  gdImageString (im_chr, gl_title,
		 (int) (0.5 * (Xsize - gl_title_len)),
		 (int) (Ysize - 1.5 * gl_title->h), gl_name[chr], fgcolor);

  if (type_disp != 1)
    {
      /* Draw the precision graphical genotype */
      for (i = 0; i < ip; i++)
	{
	  y1_pgg =
	    (i ==
	     0) ? y1_chr : (int) (y1_chr + 0.5 * scale * (p[i - 1] + p[i] -
							  2 * y_beg));
	  y2_pgg =
	    (i ==
	     (ip - 1)) ? y2_chr : (int) (y1_chr + 0.5 * scale * (p[i + 1] +
								 p[i] -
								 2 * y_beg));
	  rgb[0] = rgb[1] = rgb[2] = 0;
	  (*calcol) (r[indiv][i]);
	  pggcolor = gdImageColorAllocate (im_chr, rgb[0], rgb[1], rgb[2]);
	  gdImageFilledRectangle (im_chr, x1_chr, y1_pgg, x2_chr, y2_pgg,
				  pggcolor);
	}
    }

  /* draw the chromosome */
  gdImageRectangle (im_chr, x1_chr, y1_chr - 5, x2_chr, y2_chr + 5, fgcolor);
  for (i = mk_beg; i < mk_end; i++)
    {
      gdImageLine (im_chr, x1_mk, y_mk[i], x2_mk, y_mk[i], fgcolor);
      /* Put mk positions on the left */
      sprintf (pozit, "%1.2f", mk_pos[chr][i]);
      gdImageString (im_chr, ppf,
		     x1_mk_name - (6 * ppf->w), y_mk_name[i], pozit, fgcolor);
      /* Put mk names on the right */
      gdImageString (im_chr, ppf,
		     x2_mk_name, y_mk_name[i], mk_name[chr][i], fgcolor);

      /* draw line from bar to name */
      gdImageLine (im_chr, x2_mk, y_mk[i], (x2_mk_name - 2),
		   poz_bottom[i] + 1, fgcolor);
      /* draw line from bar to position */
      gdImageLine (im_chr, x1_mk, y_mk[i], (x1_mk_name + 2),
		   poz_bottom[i] + 1, fgcolor);

    }
  return im_chr;

}


gdImagePtr
merge_LG_images (img1, img2)
     gdImagePtr img1;
     gdImagePtr img2;

{
  gdImagePtr img_merged;
  int xsi, ysi;
  int xsi1, xsi2, ysi1, ysi2;
  int kill1, kill2;

  if (img1 != NULL)
    {
      xsi1 = img1->sx;
      ysi1 = img1->sy;
      kill1 = 0;
    }
  else
    {
      xsi1 = 0;
      ysi1 = 0;
      /* burk :( */
      img1 = ImageCreateTrueColorWhiteBG (xsi1, ysi1);
      kill1 = 1;
    }

  if (img2 != NULL)
    {
      xsi2 = img2->sx;
      ysi2 = img2->sy;
      kill2 = 0;
    }
  else
    {
      xsi2 = 0;
      ysi2 = 0;
      /* burk :( */
      img2 = ImageCreateTrueColorWhiteBG (xsi2, ysi2);
      kill2 = 1;
    }

  xsi = (xsi1 > xsi2) ? xsi1 : xsi2;
  ysi = ysi1 + ysi2 + 5;
  Ysize = (Ysize < ysi) ? ysi : Ysize;
  YSIZE = (YSIZE < Ysize) ? Ysize : YSIZE;

  img_merged = ImageCreateTrueColorWhiteBG (xsi, ysi);

  if (img_merged == NULL)
    {
      printf ("img_merged is NULL, ciao\n");
      exit (1);
    }
  if (img2 == NULL)
    {
      printf ("img2 is NULL, ciao\n");
      exit (1);
    }
  if (img1 == NULL)
    {
      printf ("img1 is NULL, ciao\n");
      exit (1);
    }

  gdImageCopy (img_merged, img1, 0, 0, 0, 0, xsi1, ysi1);

  gdImageCopy (img_merged, img2, 0, ysi1, 0, 0, xsi2, ysi2);
  if (kill1)
    gdImageDestroy (img1);
  if (kill2)
    gdImageDestroy (img2);
  return img_merged;


}

int
format_results (geno, p, prob_add, prob_mk)
     int geno, p;
     double prob_add[N];

     double prob_mk;
{
  int toplot;			/* index  of all_disp in addloc_off */
  int k;
  toplot = -1;

  switch (type_disp)
    {
    case 0:
      /* No Graphics */
      return 0;
    case 2:
      /* identify the genotype to plot */
      for (k = 0; k < nb_addloc; k++)
	toplot = (addloc_off[k] == all_disp) ? k : toplot;
      if (toplot < 0)
	{
	  fprintf (stderr, "genotype to plot = %d is unknown\n", all_disp);
	  exit (1);
	}
      /* plot the probability of genotype "toplot" */
      results[geno][p] = prob_add[toplot] / prob_mk;
      break;
    case 3:
      /* Plots the allele dose for each individual */
      results[geno][p] =
	(prob_add[0] / prob_mk) + 0.5*(prob_add[1] / prob_mk);
      break;
    case 4:
      results[geno][p] =
	(prob_add[0] / prob_mk) + 0.5*(prob_add[1] / prob_mk);
      break;
    case 5:
      results[geno][p]=-1; /* default */
      for (k = 0; k < nb_addloc; k++)
	{
	  if (threshold[addloc_off[k]].th == -1) 
	    continue;
	  else if ((prob_add[k] / prob_mk) >= threshold[addloc_off[k]].th)
	    {
	      if (!strcmp (threshold[addloc_off[k]].color, "red"))
		results[geno][p] = 0;
	      else if (!strcmp (threshold[addloc_off[k]].color, "green"))
		results[geno][p] = 1;
	      else if (!strcmp (threshold[addloc_off[k]].color, "blue"))
		results[geno][p] = 2;
	      else if (!strcmp (threshold[addloc_off[k]].color, "none"))
		results[geno][p] = -1;
	      else
		printf ("this should not arrive : argll\n");
	    }
	}
      calcol = thresh;
      break;
    }
  return 0;
}

int
main (argc, argv)
     int argc;
     char *argv[];
{
  int i, ii, j, k, l, p;
  int ipos;
  int mk_beg, mk_end;
  int chr, chr_end;
  int gl, gl_beg, gl_end;
  int chr_idx_max;
  int nb_real_chr;
  char fichier[2048];
  FILE *fd;
  char mainfile[2048];		/* name of input files */
  int temp_arg;
  double poz[POSITIONS];	/* positions scanned on each chromosome */
  gdFontPtr pf, pf_title;
  /* sub pictures of the chromosomes */
  gdImagePtr gl_pic[GENO][NGL];
  gdImagePtr chromo_pic[GENO][CHROMO];
  gdImagePtr mainpic;
  int fgcolor;
  /* output files */
  FILE *pngout, *jpegout;
  /* name of output file */
  char extension[2048];
  int XSIZE=0;
  double max_lenc, templen;
  int xmain;

  temp_arg = argc - 1;

  if (argc == 1)
    {
      fprintf (stderr,
	       "Usage : %s [options] infile \n type '%s -h' for details\n",
	       "grafgen", "grafgen");
      exit (1);
    }
  if (temp_arg < 0)
    {
      print_help (argv);
      exit (1);
    }
  options (argc, argv);
  mainfile[0] = 0;
  if (!sscanf (argv[temp_arg], "%s", mainfile))
    {
      fprintf (stderr, "invalid infile name\n");
      exit (1);
    }

  /* storing the inputs and allocating values to variables */

  chr_idx_max = reading_input (mainfile);
  gl_beg = lg[chr_beg].index[0];

  /* TEST OF INPUT */
  if (VERBOSE) {
    fprintf(stderr,"Number of chromosomes: %d\n",nb_chr);
    for (i=0; i<nb_chr;i++) {
      if (lg[i].nb>0) {
	fprintf(stderr,"Chrom %d:\n",i+1) ;
	for (j=0; j<lg[i].nb; j++) {
	  fprintf(stderr,"Linkage group %s has %d markers\n", gl_name[lg[i].index[j]],endc[lg[i].index[j]]+1);
	  fprintf(stderr,"Marker Positions:");
	  for (k=0; k<=endc[lg[i].index[j]]; k++) {
	    fprintf(stderr,"%.1f ",mk_pos[lg[i].index[j]][k]);
	  }
	  fprintf(stderr,"\n");
	}
      }
    }
  }

  if (GENOME)
    {
      gl_end = lg[nb_chr-1].index[lg[nb_chr-1].nb-1]+1;
      chr_end = nb_chr;
    }
  else
    {
      gl_end = gl_beg + lg[chr_beg].nb;
      chr_end = chr_beg + 1;
    }

  if (off_end < 0)
    off_end = nb_off;
  else if (off_end > nb_off)
    {
      fprintf (stderr, "The individual %d does not exists\n", off_end);
      exit (1);
    }
  else
    {
      nb_off = off_end;
      j = 0;
      for (i = off_beg; i < off_end; i++)
	{
	  for (gl = gl_beg; gl < gl_end; gl++)
	    {
	      if (endc[gl] < 0)
		continue;
	      for (l = 0; l < endc[gl]; l++)
		{
		  geno_off[gl][j][l] = geno_off[gl][i][l];
		}

	    }
	  j++;
	}
    }

  reading_coding_system (ficod);
  if ((type_disp == 3) || (type_disp == 4))
    {
      codsys_analysis ();
      nb_addloc = 2;
      addloc_off[0] = homo;
      addloc_off[1] = hetero;
    }
  else if (type_disp == 5)
    {
      reading_threshold (fic_thresh);
    }

  /* Initialize Constant Values */

  nb_real_chr = nb_chr;
  max_lenc = lenc[gl_beg];
  for (chr = chr_beg; chr < chr_end; chr++)
    {
      templen = 0;
      for (l = 0; l < lg[chr].nb; l++)
	{
	  templen += lenc[lg[chr].index[l]];
	}
      if (templen == 0)
	nb_real_chr--;
      max_lenc = (max_lenc < templen) ? templen : max_lenc;
    }
  if (max_lenc == 0)
    {
      fprintf (stderr, "Error: longest chromosome is of size 0\n");
      exit (1);
    }

  /* Case only here to make a nice genetic map */
  if (type_disp == 1)
    {
      fprintf(stderr,"Drawing Genetic Map\n");
      /* Initialization */
      pf = gdFontSmall;
      mainpic=NULL;
      xmain = 0;

      /* draw each chromosome in a sub-picture */
      for (gl = gl_beg; gl < gl_end; gl++)
	{
	  if (endc[gl] < 0)
	    continue;
	  if (!INTER)
	    x_end = lenc[gl];
	  if (x_beg > x_end)
	    {
	      fprintf (stderr,
		       "error on the size of the segment to map : x_beg: %.3f > x_end: %.3f\n",
		       x_beg, x_end);
	      exit (1);
	    }
	  gl_pic[0][gl] =
	    make_chromosome_image (gl, pf, x_beg, x_end, 0, NULL, NULL, NULL);
	}
      /* merge Linkage groups belonging to the same chromosome */
      for (chr = chr_beg; chr < chr_end; chr++)
	{
	  for (l = 0; l < lg[chr].nb; l++)
	    {
	      if (endc[lg[chr].index[l]] < 0)
		continue;
	      chromo_pic[0][chr] =
		merge_LG_images (chromo_pic[0][chr],
				 gl_pic[0][lg[chr].index[l]]);
	      gdImageDestroy (gl_pic[0][lg[chr].index[l]]);
	    }
	}

      /* Create the main picture */
      /* Find the piture of larger height */
      for (  Ysize = 0, chr = chr_beg; chr < chr_end; chr++) {
	if (chromo_pic[0][chr]) {
	  if (chromo_pic[0][chr]->sy > Ysize) 
	    Ysize = chromo_pic[0][chr]->sy;
	}
      }
      XSIZE = Xsize * nb_real_chr;
      YSIZE = TitleSize+Ysize;
      mainpic=ImageCreateTrueColorWhiteBG (XSIZE, YSIZE);

      /* copy sub picture of chromosomes on the main picture */
      for (chr = chr_beg; chr < chr_end; chr++)
	{
	  if (chromo_pic[0][chr] == NULL)
	    continue;
	  gdImageCopy (mainpic, chromo_pic[0][chr], xmain, TitleSize, 0, 0,
		       chromo_pic[0][chr]->sx, chromo_pic[0][chr]->sy);
	  xmain += chromo_pic[0][chr]->sx;
	}

      /* Write the Title */
      pf_title = gdFontSmall;
      fgcolor = gdImageColorAllocate (mainpic, 0, 0, 0);
      gdImageString (mainpic, pf_title, 0, (int) (TitleSize / 2), title,
		     fgcolor);
      

      /* Outputing */
      strncpy (fichier, title, 2048 * sizeof (char));
      strcat (fichier, "_map");
      if (strncmp (GFORMAT, "jpg", 3) == 0)
	{
	  strcat (fichier, ".jpg");
	  if ((jpegout = fopen (fichier, "wb")) == NULL)
	    {
	      fprintf (stderr, "Cannot open file : %s\n", fichier);
	      exit (1);
	    }
	  printf ("Output file : %s\n", fichier); 
	  gdImageJpeg (mainpic, jpegout, 90);
	  fclose (jpegout);
	}
      else if ((strncmp (GFORMAT, "png", 3) == 0)
	       || (strncmp (GFORMAT, "def", 3) == 0))
	{
	  strcat (fichier, ".png");
	  if ((pngout = fopen (fichier, "wb")) == NULL)
	    {
	      fprintf (stderr, "Cannot open file : %s\n", fichier);
	      exit (1);
	    }
	  printf ("Output file : %s\n", fichier);
	  gdImagePng (mainpic, pngout);
	  fclose (pngout);
	}
      else
	{
	  fprintf (stderr, "Graphics can only be JPEG or PNG, sorry\n");
	  exit (1);
	}
      /* Destroy the images in memory */
      for (chr = chr_beg; chr < chr_end; chr++)
	{
	  if (chromo_pic[0][chr] == NULL)
	    continue;
	  gdImageDestroy (chromo_pic[0][chr]);
	}
      gdImageDestroy (mainpic);


      return 0;
    }
  /* ELSE : compute ! */
  /* Create files and writes headers */
  strcpy (fichier, title);
  strcat (fichier, ".dat");
  if ((fd = fopen (fichier, "w")) == NULL)
    {
      fprintf (stderr, "Cannot write in file :%s\n", fichier);
      exit (1);
    }

  fprintf (fd, "chr");
  fprintf (fd, ",posx");
  for (j = off_beg; j < off_end; j++)
    {
      for (i = 0; i < nb_addloc; i++)
	{
	  fprintf(fd, ",I%dG%d", j, addloc_off[i]);
	}
    }
  fprintf (fd, "\n");
  fclose (fd);


  /* COMPUTATION OF CONDITIONAL PROBABILITIES */
  for (gl = gl_beg; gl < gl_end; gl++)
    {
      fprintf (stderr, "Analysing Linkage Group : %s : ", gl_name[gl]);
      for (i = 0; i < GENO; i++)
	{
	  for (j = 0; j < POSITIONS; j++)
	    {
	      results[i][j] = 0;
	    }
	}
      if (endc[gl] < 0)
	continue;
      if (!INTER)
	x_end = lenc[gl];
      ipos = 0;
      if (x_beg > x_end)
	{
	  fprintf (stderr,
		   "error on scanning positions : x_beg: %.3f > x_end: %.3f\n",
		   x_beg, x_end);
	  exit (1);
	}
      if ((x_end - x_beg) / xstep > POSITIONS)
	{
	  fprintf (stderr,
		   "Too many positions to scan : %d > MAX POSITIONS = %d\n",
		   (int) (1 + (x_end - x_beg) / xstep), POSITIONS);
	  exit (1);
	}
      for (posx = x_beg; posx <= x_end; posx += xstep)
	{
	  poz[ipos] = posx;
	  /* markers to be taken in to account */

	  /* flanking markers */
	  for (i = 0; i < endc[gl] && mk_pos[gl][i] <= posx; i++);
	  mk_beg = i - 1;
	  mk_end = i;

	  /* adapting mk_beg and mk_end to mstep */
	  if (mk_beg < 0)
	    mk_beg = mk_end;

	  if ((mk_beg - mstep) >= 0)
	    mk_beg -= mstep;
	  else
	    mk_beg = 0;

	  if ((mk_end + mstep) < endc[gl])
	    mk_end += mstep;
	  else
	    mk_end = endc[gl] - 1;

	  mdm_calc (mk_beg, mk_end, endc[gl], mk_pos[gl],
		    gen, geno_mat[gl], mating_system, geno_pat[gl], nb_off,
		    geno_off[gl], 0, 0, addloc_mat, addloc_pat, addloc_off,
		    probcomm_mq, probdiff_mq);

	  mdm_calc (mk_beg, mk_end, endc[gl], mk_pos[gl],
		    gen, geno_mat[gl], mating_system, geno_pat[gl], nb_off,
		    geno_off[gl], nb_addloc, posx, addloc_mat, addloc_pat,
		    addloc_off, probcomm, probdiff);

	  /* Write condit. probs in dat files */
	  strcpy (fichier, title);
	  strcat (fichier, ".dat");
	  if ((fd = fopen (fichier, "a")) == NULL)
	    {
	      fprintf (stderr, "Cannot write in file :%s\n", fichier);
	      exit (1);
	    }
	  fprintf(fd,"%s",gl_name[gl]);
	  fprintf (fd,",%3.1f", poz[ipos]);
	  for (i = off_beg; i < off_end; i++)
	    {
	      for (k = 0; k < nb_addloc; k++)
		{
		  if (probdiff_mq[i][0] == 0)
		    {
		      fprintf (stderr,
			       "\nImpossible genotype identified !\nPlease check your data\n");
		      fprintf (stderr, "Individual %d, position %f\n", i,
			       posx);
		      exit (1);
		    }
		  else
		    {
		      fprintf(fd, ",%1.5f",probdiff[i][k] / probdiff_mq[i][0]);
		    }
		}
	      /* Adapt Results to what we want */
	      format_results (i, ipos, probdiff[i], probdiff_mq[i][0]);
	    }
	  fprintf (fd, "\n");
	  fclose (fd);
	  ipos++;
	}
      
      if (type_disp == 4)
	{
	  if (expect_val > -1)
	    {
	      /* We perform a simple test and plot a "p-value" */
	      /* Student t-test */
	      double dmean, dvar, *ttest;
	      int ssize;
	      ttest = malloc (ipos * sizeof (double));
	      if (ttest == NULL)
		{
		  fprintf (stderr, "Alloc error\n");
		  exit (1);
		}
	      ssize = off_end - off_beg;
	      if (ssize < 30)
		fprintf (stderr,
			 "WARNING: %d individuals is not enough to test for deviation\n",
			 ssize);

	      for (p = 0; p < ipos; p++)
		{
		  dmean = 0;
		  dvar = 0;
		  for (i = off_beg; i < off_end; i++)
		    {
		      dmean += results[i][p];
		      dvar += results[i][p] * results[i][p];
		    }
		  dmean /= ssize;
		  dvar /= ssize;
		  dvar = dvar - (dmean * dmean);

		  ttest[p] = (dmean - expect_val) / sqrt (dvar / ssize);
		}
	      free (ttest);
	      /* !!! We overwrite the first individual values to store the population p-values */
	      for (p = 0; p < ipos; p++)
		{
		  results[0][p] = ttest_pval (ttest[p]);
		}
	    }
	  else
	    {
	      /* We plot the mean allele dose */
	      double dmean;
	      int ssize;
	      ssize = off_end - off_beg;
	      for (p = 0; p < ipos; p++)
		{
		  dmean = 0;
		  /* mean dose at this position */
		  for (i = off_beg; i < off_end; i++)
		    dmean += results[i][p];
		  dmean /= ssize;
		  results[0][p]=dmean;
		}
	    }
	  pf = gdFontSmall;
	  gl_pic[0][gl] =
	    make_chromosome_image (gl, pf, x_beg, x_end, 0, poz, results,
				   ipos);
	}
      /* If image files are wanted */
      else if (type_disp)
	{
	  /* create images of each chromosome */
	  for (ii = off_beg; ii < off_end; ii++)
	    {
	      pf = gdFontSmall;
	      gl_pic[ii][gl] =
		make_chromosome_image (gl, pf, x_beg, x_end, ii, poz, results,
				       ipos);
	      assert(gl_pic[ii][gl]);
	    }
	}
      fprintf (stderr, "done\n");
    }
  
  /* If images files are wanted */
  if (type_disp)
    {
      if (type_disp == 4)
	{
	  off_beg = 0;
	  off_end = 1;
	}
      if (GENOME) {
	YSIZE=Ysize+TitleSize;
	XSIZE=Xsize*nb_real_chr;
      }
      for (ii = off_beg; ii < off_end; ii++)
	{
	  pf = gdFontSmall;
	  /* merge Linkage groups belonging to the same chromosome */
	  for (chr = chr_beg; chr < chr_end; chr++)
	    {
	      for (l = 0; l < lg[chr].nb; l++)
		{
		  if (endc[lg[chr].index[l]] < 0)
		    continue;
		  chromo_pic[ii][chr] =
		    merge_LG_images (chromo_pic[ii][chr],
				     gl_pic[ii][lg[chr].index[l]]);
		  gdImageDestroy (gl_pic[ii][lg[chr].index[l]]);
		}
	    }
	  if (!GENOME)
	    {
	      /* create main picture */
	      mainpic = ImageCreateTrueColorWhiteBG (Xsize, Ysize);
	      assert(mainpic);
	    }
	  else
	    {
	      pf_title = gdFontSmall;
	      /* create mainpic */
	      mainpic = ImageCreateTrueColorWhiteBG (XSIZE, YSIZE);
	      /* Write Title, logo ? */
	      fgcolor = gdImageColorAllocate (mainpic, 0, 0, 0);
	      gdImageString (mainpic, pf_title, 0, TitleSize / 2, title,
			     fgcolor);
	    }
	  /* draw main */
	  xmain = 0;
	  /* copy sub picture of chromosomes on the main picture */
	  for (chr = chr_beg; chr < chr_end; chr++)
	    {
	      if (chromo_pic[ii][chr] == NULL)
		continue;
	      gdImageCopy (mainpic, chromo_pic[ii][chr], xmain, TitleSize, 0, 0,
			   chromo_pic[ii][chr]->sx, chromo_pic[ii][chr]->sy);
	      xmain += chromo_pic[ii][chr]->sx;
	      gdImageDestroy (chromo_pic[ii][chr]);
	    }


	  strcpy (fichier, title);
	  sprintf (extension, "_i%d", (ii + 1));
	  strcat (fichier, extension);
	  /* Open files for PNG or JPEG out  and copy picture */
	  if (strncmp (GFORMAT, "jpg", 3) == 0)
	    {
	      strcat (fichier, ".jpg");
	      if ((jpegout = fopen (fichier, "wb")) == NULL)
		{
		  fprintf (stderr, "Cannot open file : %s\n", fichier);
		  exit (1);
		}
	      printf ("Output file : %s\n", fichier);
	      assert(mainpic);
	      gdImageJpeg (mainpic, jpegout, 90);
	      fclose (jpegout);
	    }
	  else if ((strncmp (GFORMAT, "png", 3) == 0)
		   || (strncmp (GFORMAT, "def", 3) == 0))
	    {
	      strcat (fichier, ".png");
	      if ((pngout = fopen (fichier, "wb")) == NULL)
		{
		  fprintf (stderr, "Cannot open file : %s\n", fichier);
		  exit (1);
		}
	      printf ("Output file : %s\n", fichier);
	      assert(mainpic);
	      gdImagePng (mainpic, pngout);
	      fclose (pngout);
	    }
	  gdImageDestroy (mainpic);
	}
    }
  return 0;
}
