/* grafgen : Precision graphical genotypes based on marker informations */
/* Copyright 2002 Bertrand Servin, Christine Dillmann, Guy Decoux and Frederic Hospital */
/* Institution :
   Station de Genetique Vegetale INRA / UPS / INAPG 
   91190 Gif-sur-Yvette - FRANCE */

/*     This file is part of Grafgen. */

/*     Grafgen is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or */
/*     (at your option) any later version. */

/*     Grafgen is distributed in the hope that it will be useful, */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*     GNU General Public License for more details. */

/*     You should have received a copy of the GNU General Public License */
/*     along with Grafgen; if not, write to the Free Software */
/*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Please report Bugs and Comments to servin@moulon.inra.fr */
   

#define MDM_CALC 1
#include "grafgen.h"
#undef MDM_CALC

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

int ijtot;

/* The equations number below refer to Appendix A of Hospital F,
   Dillmann C, and Melchinger AE, 1996, A general Algorithm to compute
   multilocus genotype frequencies under various mating systems,
   Comput. App. Biosci. 12 : 455--462 */

#define LOCOL 7
#define IJ (1 << LOCOL)

double *Rk = 0;          /* Probabilities of each recombination
                            event. Equation A.6 */

double pij[IJ][IJ][IJ];  /* Probabilities for every TG to produce
                            every gamete.  Equation A.8 */

double Fij[3][IJ][IJ];   /* Theoretical frequencies of TGs at
                            generations (j-1) and j (first two
                            dimensions) and of the second parent if
                            exists (third dimension) */

int DPres[3][IJ * IJ + 1][2];    /* TGs corresponding to the OGs of
                                    the maternal ancestors at
                                    generations (j-1) and j (first two
                                    dimensions) and of the paternal
                                    ancestors if exists (third
                                    dimension) */

/* The functions zu and uz allow to associate bijectively integers
   with tables filled following a given base */
/* Typically, if the base is 2, the tables will be the binary code of 
   the number (they will be filled with 0 and 1).
   If the base is 3, the tables will be filled with 0, 1 and 2 
   and so on ...*/

/* Function zu : Equation A.1 or A.4 depending on the value of 'nball' */
/* decodes a gamete defined as an integer *n* into a table (*tab*) of size */
/* *locus_tot* following base *nball* */
/* N.B. : this function is also used to decode multilocus recombination */
/* events (base 2) - see below, e.g. in function *Kc* -*/

#define zu(tab, nball, n, locus_tot)			\
{						\
    int i00, j00;				\
    j00 = n;					\
    for (i00 = 0; i00 < locus_tot; i00++)	\
        tab[i00] = 0;				\
    for (i00 = 0; j00; j00 /= nball, i00++) {	\
        tab[i00] = j00 % nball;			\
    }						\
}


/* Function uz : Equations A.2 */
/* codes a gamete defined by a table (*tab*) into an integer (*res*)  */ 
/* following the base *nball*. */
 
int uz(tab, n)
    int *tab, n;
{
    int i, j, res;
    res = 0;
    for (i = 0, j = 1; i < n; i++, j *= nball)
        res += tab[i] * j;
    return res;
}


/* Kc: computes the probability associated to each possible series of
   recombination events between loci, depending on the recombination
   rates between loci, Equation A.6 */
/* These probabilities are stored in the table Rk */
/* If two adjacent loci (i and i+1) issue from two different parents in a given 
   recombination event, the probability associated with this 
   recombination event is multiplied by rfact[i] (the recombination 
   rate between the two loci) if they issue from the same parent, it is 
   multiplied by (1-rfact[i]) */

void Kc(rfact, locus_tot)
    double *rfact;
    int locus_tot;
{
    int tab[LOCOL], i, n, nbrec;
    double result;

    nbrec = 1;

    /* Computing the number of recombination events - 2^locus_tot. */

    for (n = 0; n < locus_tot; n++)
	nbrec *= 2;

    /* For each recombination event 'n' ... */
    for (n = 0; n < nbrec; n++) {
	/* ... decoding the corresponding table 'tab'  */
	/* giving parental origin of allele at each locus.*/ 
        zu(tab, 2, n, locus_tot);
        result = 1;
	
	/* For each couple of adjacent loci (i,i+1) ...*/
        for (i = 0; i < locus_tot - 1; i++) {
	    /* ... times result by (1-rfact[i]) if tab[i]=tab[i+1] - allele of the same 
	       parental origin -, else by rfact[i]. */
            result *= (tab[i] == tab[i + 1])?(1 - rfact[i]):rfact[i];
        }

	result /= 2;
        Rk[n] = result;
    }
}


/* pijc: computes the probability pij[k] that a TG (i,j) produces by
   meiosis the gamete k. Equation A.8 */

static void
pijc(ntot, locus_tot)
    int ntot, locus_tot;
{
    int git[2][LOCOL], res[LOCOL], tab[LOCOL];
    int n, i, gi, gj, nbrec;

    nbrec = 1;

    /* Computing the number of recombination events 'nbrec' = 2^locus_tot */
    for (n = 0; n < locus_tot; n++)
	nbrec *= 2;

    /* Allocating memory to table pij and initializing values to 0*/
    memset(pij, 0, IJ * IJ * IJ * sizeof(double));

    /* For each possible couple of gametes (gi, gj) - integers -
       the gametes are decoded into tables git[0] and git[1]. */
    
    for (gi = 0; gi < ntot; gi++) {
	zu(git[0], nball, gi, locus_tot);
	for (gj = 0; gj < ntot; gj++) {
	    zu(git[1], nball, gj, locus_tot);
	    
	    /* Each recombination event 'n' is applied to the couple (git[0],git[1]) 
	       and produces the gamete 'res'. */
	    for (n = 0; n < nbrec; n++) {
		zu(tab, 2, n, locus_tot);
		for (i = 0; i < locus_tot; i++)		    
		    res[i] = git[tab[i]][i];
		/* 'res' is coded into an integer and its frequency is added 'Rk[n]',
		   the probability of the recombination event. */
		pij[gi][gj][uz(res, locus_tot)] += Rk[n];
	    }
	}
    }
}


/* Remplissage: computes DPres values. */
/*     - Identifies the couples of gametes (i2,j2) of TGs corresponding to */
/*     an OG ('tabi') at generation (j-1) or j (indice = 0 or 1), or of the */
/*     paternal genotype ('indice' = 2)  */
/*     - Computes, in a similar way as the function 'uz' above, the codes of */
/*     these gametes, following the informations of the structure coding_system  */


static int
remplissage(total, indice, tabi, geni, i1, j1, locus_tot)
    int total, indice, geni, i1, j1;
    int *tabi;
{
    int vali, i, geno;
    int i2, j2;

    geno = 1;
    for (i = 0; i < geni; i++)
	geno *= nball;
    vali = tabi[geni];
    for (i = 0; i < coding_system[vali].len; i++) {
	i2 = i1 + coding_system[vali].possible[i][0] * geno;
	j2 = j1 + coding_system[vali].possible[i][1] * geno;
	if (geni == (locus_tot - 1)) {
	    DPres[indice][total][0] = i2;
	    DPres[indice][total][1] = j2;
	    total++;
	}
	else {
	    total = remplissage(total, indice, tabi, geni + 1, i2, j2, locus_tot);
	}
    }
    return total;
}


/* remplis: calls the function 'remplissage' above to identify the TGs
   associated to a OG */

static int
remplis(indice, tabi, locus_tot)
    int indice, locus_tot;
    int *tabi;
{
    int total, geni, i1, j1;
    total = geni = i1 = j1 = 0;
    total = remplissage(total, indice, tabi, geni, i1, j1, locus_tot);
    DPres[indice][total][0] = DPres[indice][total][1] = -1;
    return total;
}

/* The functions 'hybride', 'backcross', 'self' and 'fullsib' are the 
   functions that allow to compute the frequencies of all TGs(g), knowing 
   the frequencies of the TGs(g-1), using the recurrences defined by 
   Hospital et al. (1996) */
/* The indice i0 defines the maternal population at generation (g-1) 
   and i1 the maternal population at generation (g). 
   If a paternal population is used, the indice to define it is always 2 */

/* Doubled Haploids, note in Hospital et al 1997 */

static double
hd(i0,i1)

   int i0, i1;
{
	
	int m, l;
	int i, j, x, y;
	double fin = 0;
       /* The initial frequencies of population i1 are set to 0 */
	memset(Fij[i1], 0, IJ * IJ * sizeof(double));

	for (m = 0; m < ijtot; m++) {
		if (DPres[i1][m][0] == -1)
			break;
		i = DPres[i1][m][0];
		j = DPres[i1][m][1];
		/* double haploids are necessarily completely homozygous */
		if ( i != j ) {
			Fij[i1][i][j] = 0;
			continue;
		}
		if (Fij[i1][i][j])
			continue;
		for (l = 0; l < ijtot; l++) {
			if (DPres[i0][l][0] == -1)
				break;
			x = DPres[i0][l][0];
			y = DPres[i0][l][1];
			if (Fij[i0][x][y] == 0)
				continue;
			Fij[i1][i][j] += pij[x][y][i]*Fij[i0][x][y];
		}
		fin += Fij[i1][i][j];
	}
	return fin;
}	

static double
hybride(i0, i1) 

/* Equation (3) : hybride computes frequencies in the case of hybrid mating */

    int i0, i1;
{
    int k, l, m;
    int u, v, x, y, i, j;
    double fin = 0;
    
    /* The initial frequencies of population i1 are set to 0 */
    memset(Fij[i1], 0, IJ * IJ * sizeof(double)); 
    
    /* For every possible genotype (TG) of the population i1 ...*/
    for (m = 0; m < ijtot; m++) {
	if (DPres[i1][m][0] == -1)
	    break;
	/* ... identifying the corresponding couple of gametes (i,j). */
	i = DPres[i1][m][0];
	j = DPres[i1][m][1];
	if (Fij[i1][i][j])
	    continue;
	/* For every possible genotype (TG) of the population i0 ...*/
	for (l = 0; l < ijtot; l++) {
	    if (DPres[i0][l][0] == -1)
		break;
	    /* ... identifying the corresponding couple of gametes (x,y). */
	    x = DPres[i0][l][0];
	    y = DPres[i0][l][1];
	    if (Fij[i0][x][y] == 0)
		continue;
	    /* For every possible genotype (TG) of the paternal population  ...*/
	    for (k = 0; k < ijtot; k++) {
		if (DPres[2][k][0] == -1)
		    break;
		/* ... identifying the corresponding couple of gametes (u,v). */
		u = DPres[2][k][0];
		v = DPres[2][k][1];
		if (Fij[2][u][v] == 0)
		    continue;
		/* The frequency of individual (i,j) in population i1 is added the 
		   probability of being produced by the mating of (u,v) by (x,y), 
		   according to their frequencies in the populations i0 and '2', 
		   respectively*/
		Fij[i1][i][j] += pij[x][y][i] * pij[u][v][j] * Fij[i0][x][y] * Fij[2][u][v];
	    }
	}
	fin += Fij[i1][i][j]; /* equals 1 at the end of the computation */
    }
    return fin;
}


static double
backcross(i0, i1, parent, locus_tot)
    int i0, i1, locus_tot;
    int *parent;

/* 'backcross' defines the maternal population and the paternal population
   as two different populations, which were input by the user.*/

{
    int j, k;
    double fin;

    memset(Fij[2], 0, IJ * IJ * sizeof(double)); 
    j = remplis(2 , parent, locus_tot);
    /* If the OG of the paternal ancestor corresponds to more than one TG,
       every TG is assumed to have the same frequency in the population */
    for (k = 0; k < ijtot; k++) {
	if (DPres[2][k][0] == -1)
	    break;
	Fij[2][DPres[2][k][0]][DPres[2][k][1]] = 1 / (double)j;
    }
    
    /* Once the populations are defined, 'backcross' calls the function 'hybride'. */
    
    fin = hybride(i0, i1);
    return fin;
}

static double
fullsib(i0, i1)
    int i0, i1;

/* 'fullsib' defines the paternal population to be the same as the maternal population*/
/* i.e. DPres[2] = DPres[i0] ... */

{
    int k;
    double fin;

    memcpy(Fij[2], Fij[i0], IJ * IJ * sizeof(double)); 
    for (k = 0; k < ijtot; k++) {
	DPres[2][k][0] = DPres[i0][k][0];
	DPres[2][k][1] = DPres[i0][k][1];
	if (DPres[2][k][0] == -1)
	    break;
    }

    /* Once the populations are defined, 'backcross' calls the function 'hybride'. */

    fin = hybride(i0, i1);
    return fin;
}


static double
self(i0, i1) /* Equation (10)  : 'self' computes frequencies in the case of selfing */
    int i0, i1;
{
    int i, j, x, y, k, l;
    double fin = 0;

    /* The initial frequencies of population i1 are set to 0 */
    memset(Fij[i1], 0, IJ * IJ * sizeof(double)); 
    
    /* For every possible genotype (TG) of the population i1 ...*/
    for (l = 0; l < ijtot; l++) {
	if (DPres[i1][l][0] == -1)
	    break;
	/* ... identifying the corresponding couple of gametes (i,j). */
	i = DPres[i1][l][0];
	j = DPres[i1][l][1];
	if (Fij[i1][i][j])
	    continue;
	/* For every possible genotype (TG) of the population i0 ...*/
	for (k = 0; k < ijtot; k++) {
	    if (DPres[i0][k][0] == -1)
		break;
	    /* ... identifying the corresponding couple of gametes (x,y). */
	    x = DPres[i0][k][0];
	    y = DPres[i0][k][1];
	    /* The frequency of individual (i,j) in population i1 is added the 
	       probability of being produced by the mating of (x,y) by itself
	       - (x,y)*(x,y) - according to its frequency in population i0 */
	    Fij[i1][i][j] += pij[x][y][i] * pij[x][y][j] * Fij[i0][x][y];
	}
	fin += Fij[i1][i][j];
    }
    return fin;
}


/* distance: computes the recombination rates from map distances using
   Haldane mapping function */

void distance(locus_tot, position, rfact)
    int locus_tot;
    double *position;
    double *rfact;
{
    int i;
    for (i = 0; i < (locus_tot - 1); i++) {
        rfact[i] = 0.5 * (1 - exp((-2 * position[i]) / 100));
    }
}


/* core computations */

void
mdm_calc(locdeb, locfin, locus_max, position, 
	 ncomm, genocomm, plans, genopar, ndiff, genodiff,
	 nx, posx, xcomm, xpar, xdiff,
	 probcomm, probdiff)
     int locdeb, locfin, locus_max;
     double position[LOCUS + 1];
     int ncomm, ndiff, nx;
     int genocomm[GENO][LOCUS + 1], genodiff[GENO][LOCUS + 1],
	 genopar[GENO][LOCUS + 1];
     char plans[GENO][256];
     double posx;
     int xcomm[GENO], xpar[GENO], xdiff[GENO];
     double probcomm[GENO], probdiff[GENO][N];
{
    int locus_tot, ntot;
    int indice, i, j, ic, intercal;
    double rfact[LOCOL - 1];
    double temppos[LOCOL];
    int tempcomm[GENO][LOCOL];
    int tempdiff[GENO][LOCOL];
    int tempment[GENO][LOCOL];

    if (locdeb < 0 || locfin > locus_max || locdeb > locfin) {
        fprintf(stderr, "mdm_calc: error on loc_beg or loc_end\n");
        exit(1);
    }
    
    locus_tot = locfin - locdeb + 1; /* number of loci o be taken into account */
    
    if (locus_tot > LOCOL) {
        fprintf(stderr, "mdm_calc: error too many loci\n");
        exit(1);
    }
    
    /* resizing tables according to locbeg ('locdeb') and locend ('locfin') */
    for (j = 0; j < locus_tot; j++)
	temppos[j] = position[locdeb + j];
    for (i = 0; i < ncomm; i++) 
	for (j = 0; j < locus_tot; j++) {
	    tempcomm[i][j] = genocomm[i][locdeb + j];
	    tempment[i][j] = genopar[i][locdeb + j];
	}
    for (i = 0; i < ndiff; i++)
	for (j = 0; j < locus_tot; j++) {
	    tempdiff[i][j] = genodiff[i][locdeb + j];
	}
    
    /* If an additional locus is to be taken into account ... */
    if (nx > 0) {
	intercal = 1;
	ic = locus_tot;
    }
    else {
	intercal = 0;
	nx = 1;
	ic = -1;
    }
    
    /* ... including the locus in the genotypes (maternal, paternal, and offspring) */
    if (intercal) {
	for (i = 0; i < locus_tot; i++)
	    if (posx <= temppos[i]) {
		ic = i;
		break;
	    }
	if (ic != locus_tot) {
	    memmove(&temppos[ic + 1], &temppos[ic],
		    (locus_tot - ic) * sizeof(double));
	    for (i = 0; i < ncomm; i++) {
		memmove(&tempcomm[i][ic + 1], &tempcomm[i][ic],
			(locus_tot - ic) * sizeof(int));
		memmove(&tempment[i][ic + 1], &tempment[i][ic],
			(locus_tot - ic) * sizeof(int));
	    }
	    for (i = 0; i < ndiff; i++) {
		memmove(&tempdiff[i][ic + 1], &tempdiff[i][ic],
			(locus_tot - ic) * sizeof(int));
	    }
	}
	temppos[ic] = posx;
	for (i = 0; i < ncomm; i++) {
	    tempcomm[i][ic] = xcomm[i];
	    tempment[i][ic] = xpar[i];
	}
	for (i = 0; i < ndiff; i++) {
	    tempdiff[i][ic] = xdiff[0];
	}
	locus_tot++;
    }
    
    if (locus_tot > LOCOL) {
        fprintf(stderr, "mdm_calc: error too many loci\n");
        exit(1);
    }


    /* Allocating memory for table Rk */
    ntot = 1;
    for (i = 0; i < locus_tot; i++)
	ntot *= nball;
    ijtot = ntot * ntot + 1;
    if ((Rk = (double *)malloc(ntot * sizeof(double))) == NULL) {
        fprintf(stderr, "mdm_calc : not enough memory\n");
        exit(1);
    }
    
    /* Computing relative distance between adjacent loci */
    for (i = 0; i < (locus_tot - 1); i++)
	temppos[i] = temppos[i + 1] - temppos[i];
    
    /* Computing recombination rates between loci */
    distance(locus_tot, temppos, rfact);
    
    /* Computing probabilites of multilocus recombination events, according to
       recombination rates between loci ('rfact') */
    Kc(rfact, locus_tot);
    
    /* Computing the probabilities of transition between every TG to 
      every gamete by meiosis */
    pijc(ntot, locus_tot);
    
    /* Initializing Frequencies of the first generation :
       if the OG of the maternal population at first generation corresponds
       to many TGs, all the TGs are assumed to have the same frequency in 
       the population */
    memset(Fij[0], 0, IJ * IJ * sizeof(double)); 
    indice = 0;
    j = remplis(indice, tempcomm[0], locus_tot);
    for (i = 0; i < ijtot; i++) {
	if (DPres[indice][i][0] == -1)
	    break;
	Fij[indice][DPres[indice][i][0]][DPres[indice][i][1]] = 1 / (double)j;
    }
    
    /* Initializing the tables containing the final results */
    for (i = 0; i < ncomm; i++)
	probcomm[i] = 1;
    for (i = 0; i < ndiff; i++) {
	for (j = 0; j < nx; j++) {
	    probdiff[i][j] = 1;
	}
    }
    
    /*For each generation of ancestors ...  */
    for (i = 1; i < ncomm; i++) {
	/* ... identifying the corresponding TG to the OG ... */
	remplis(1 - indice, tempcomm[i], locus_tot);
	/* ... identifying the mating system and mating individuals according to it. */
	if (!strncmp(plans[i - 1], "bc", 2) ||
	    !strncmp(plans[i - 1], "hyb", 3)) {
	    probcomm[i] *=
		backcross(indice, 1 - indice, tempment[i - 1], locus_tot);
	}
	else if (!strncmp(plans[i - 1], "self", 4)) {
	    probcomm[i] *= self(indice, 1 - indice);
	}
	else if (!strncmp(plans[i - 1], "fs", 2)) {
	    probcomm[i] *= fullsib(indice, 1 - indice);
	}
	else if (!strncmp(plans[i - 1], "hd", 2)) {
		probcomm[i] *= hd(indice, 1 - indice);
	}	
	else {
	    fprintf(stderr, "invalid mating system: %s at generation %d\n", plans[i - 1], i);
	    exit(1);
	}
	/* Passing to the next generation. */
	indice = 1 - indice;
    }
    /* N.B. : the results of these computations are stored in 'probcomm'. */
    
    /* Doing the same as before for the last generation (offspring), 
       the results are stored in 'probdiff' */
    for (i = 0; i < ndiff; i++) {
	for (j = 0; j < nx; j++) {
	    if (intercal)
		tempdiff[i][ic] = xdiff[j];
	    remplis(1 - indice, tempdiff[i], locus_tot);
	    if (!strncmp(plans[ncomm - 1], "bc", 2) ||
		!strncmp(plans[ncomm - 1], "hyb", 3)) {
		probdiff[i][j] *= backcross(indice, 1 - indice,
					    tempment[ncomm - 1], locus_tot);
	    }
	    else if (!strncmp(plans[ncomm - 1], "self", 4)) {
		probdiff[i][j] *= self(indice, 1 - indice);
	    }
	    else if (!strncmp(plans[ncomm - 1], "fs", 2)) {
		probdiff[i][j] *= fullsib(indice, 1 - indice);
	    }
	    else if (!strncmp(plans[ncomm - 1], "hd", 2)) {
		    probdiff[i][j] *= hd(indice, 1 - indice);
	    }	
	    else {
		fprintf(stderr, "invalid mating system: %s at generation %d\n", plans[i - 1], ncomm);
		exit(1);
	    }
	}
    }
}
